clear all
close all
clc

%% LDPC code design parameters

% Noise variance (set to 1 without loss of generality)
sigma_2 = 1;

% User powers (equivalently, SNRs since sigma_2 = 1)
P1 = 1.5;
P2 = 1;

% Check node degree distributions (concentrated)
dc1 = 8;
dc2 = 7;

% Maximum variable node degree
v_max1 = 100;
v_max2 = 100;

% Maximum number of alternating optimization iterations
maxiter = 3;

% Create EXIT chart axes
% Maximum value 0.99 to avoid numerical problems. Stability condition takes
% care of convergence.
step = 1e-4;
Iav = 0.001:step:0.99;
Iec = Iav;

fprintf(1, 'Designing LDPC codes for the following parameters:\n');
fprintf(1, '------------------------------\n');
fprintf(1, '   User 1    \t   User 2   \n');
fprintf(1, '------------------------------\n');
fprintf(1, 'P1 = %.2f \t\tP2 = %.2f\n', P1, P2);
fprintf(1, 'dc1 = %d \t\tdc2 = %d\n', dc1, dc2);
fprintf(1, 'vmax1 = %d \t\tvmax2 = %d\n', v_max1, v_max2);
fprintf(1, '------------------------------\n');
fprintf(1, 'Maximum iterations = %d\n', maxiter)
fprintf(1, '------------------------------\n');


%% Pre-processing and initialization

% Create check node EXIT charts
Iac1 = 1 - Jfunc( sqrt( (1 ./ (dc1-1)) * Jinv(1 - Iec).^2 ) );
Iac2 = 1 - Jfunc( sqrt( (1 ./ (dc2-1)) * Jinv(1 - Iec).^2 ) );

% Get coefficients for rate calculation
coeff1 = 2:v_max1;
coeff2 = 2:v_max2;

% Initial distribution for user 1 (has very little effect on the alternating optimization result)
lambda1 = zeros(1, v_max1-1);
lambda1(1) = 0.7;
lambda1(end) = 0.3;
lambda1 = lambda1 / sum(lambda1);

% Calculate initial rate for user 1
rate1 = 1 - (1/dc1) / sum(lambda1./coeff1);

% Calculate initial node perspective degree distribution
L1 = lambda1 ./ (2:v_max1);
L1 = L1 / sum(L1);

% Store rates for each iteration
rate1_iter = zeros(1, maxiter+1);
rate1_iter(1) = rate1;
rate2_iter = zeros(1, maxiter+1);
sumrate_iter = zeros(1, maxiter+1);
sumrate_iter(1) = rate1;

%% Perform alternating optimization

% Set cvx to quiet mode
cvx_quiet(true);

for jj = 1:maxiter
    
    % Print information
    fprintf(1, '--- Iteration %d ---\n', jj);

    % Calculate variable to state mutual information for user 1
    Ievs1 = zeros(size(Iav));
    for i = 2:v_max2
        Ievs1 = Ievs1 + L1(i-1) * Jfunc(sqrt(i)*Jinv(Iav));
    end

    % Calculate EXIT chart for each variable degree of user 2
    Iev2 = zeros(v_max2-1, length(Iav));
    precalc = Jinv( 0.5*Jfunc(sqrt(2*F2func11(Jinv( Ievs1 ).^2/2, P1, P2))) + 0.5*Jfunc(sqrt(2*abs(F2func1m1(Jinv( Ievs1 ).^2/2, P1, P2))))).^2;
    for j = 2:v_max2
        Iev2(j-1,:) = Jfunc(sqrt(precalc + (j-1)*Jinv(Iav).^2 ));
    end

    % Maximize the rate for user 2
    fprintf(1, 'Maximizing rate for user 2...');
    cvx_begin
        variable lambda2(1,v_max2-1);  
        maximize sum(lambda2./coeff2)
        subject to
            % EXIT chart condition
            Iac2 <= lambda2 * Iev2;
            % Degree distribution conditions
            sum(lambda2) == 1;
            lambda2 >= 0;
            % Stability condition
            lambda2(1) <= exp(P2/(2*sigma_2))/(dc2-1);
    cvx_end
    fprintf(1, ' done!\n');    

    % Calculate rate of user 2
    rate2 = 1 - (1/dc2) / cvx_optval;

    % Calculate node perspective degree distribution for user 2
    L2 = lambda2 ./ coeff2;
    L2 = L2 / sum(L2);

    % Calculate variable to state mutual information for user 2
    Ievs2 = zeros(size(Iav));
    for i = 2:v_max2
        Ievs2 = Ievs2 + L2(i-1) * Jfunc(sqrt(i)*Jinv(Iav));
    end

    % Calculate EXIT chart for each variable degree of user 1
    Iev1 = zeros(v_max1-1, length(Iav));
    precalc = Jinv( 0.5*Jfunc(sqrt(2*F1func11(Jinv( Ievs2 ).^2/2, P1, P2))) + 0.5*Jfunc(sqrt(2*F1func1m1(Jinv( Ievs2 ).^2/2, P1, P2)))).^2;
    for j = 2:v_max1
        Iev1(j-1,:) = Jfunc(sqrt(precalc + (j-1)*Jinv(Iav).^2 ));
    end

    % Maximize the rate for user 1
    fprintf(1, 'Maximizing rate for user 1...');
    cvx_begin
        variable lambda1(1,v_max1-1);  
        maximize sum(lambda1./coeff1)
        subject to
            % EXIT chart condition
            Iac1 <= lambda1 * Iev1;
            % Degree distribution conditions
            sum(lambda1) == 1;
            lambda1 >= 0;
            % Stability condition
            lambda1(1) <= exp(P1/(2*sigma_2))/(dc1-1);
    cvx_end
    fprintf(1, ' done!\n');  

    % Calculate rate of user 1
    rate1 = 1 - (1/dc1) / cvx_optval;

    % Output and store rates for current iteration
    fprintf('Rate 1: %.3f bits/ch. use\nRate 2: %.3f bits/ch. use\nSum Rate: %.3f bits/ch. use\n', rate1, rate2, rate1 + rate2);
    rate1_iter(jj+1) = rate1;
    rate2_iter(jj+1) = rate2;
    sumrate_iter(jj+1) = rate1 + rate2;

    % Calculate node perspective degree distribution for user 1
    L1 = lambda1 ./ coeff1;
    L1 = L1 / sum(L1);

end

%% Plot results

% Plot rate vs iterations
figure(1)
plot(1:maxiter+1, rate1_iter(1:end), 'o-', 'LineWidth',2)
hold on
plot(1:maxiter+1, rate2_iter(1:end), 's-', 'LineWidth',2)
plot(1:maxiter+1, sumrate_iter(1:end), '^-', 'LineWidth',2)
xlabel('Iterations')
ylabel('Rate')
legend('User 1 Rate', 'User 2 Rate', 'Sum Rate', 'location', 'best')
set(gca,'XTick',1:maxiter+1)
grid on

% Plot capacity region and solution
[ c1, c2, c12 ] = MACapacity(P1, P2, 1, 1e7);
fprintf('The maximal sum-rate for P1 = %.2f and P2 = %.2f is: %.3f bits/ch. use\n', P1, P2, c12)
step = 0.0001;
x_axis = 0:step:1;
c1_axis = 0:step:c2;
figure(2)
plot(rate2, rate1, 'rx', 'LineWidth', 2, 'MarkerSize', 9)
hold on
line(x_axis(1:ceil((c12-c1)/step)), c1*ones(1,length(x_axis(1:ceil((c12-c1)/step)))),'LineWidth',2)
line(c2*ones(1,length(x_axis(1:ceil((c12-c2)/step)))), x_axis(1:ceil((c12-c2)/step)),'LineWidth',2)
line(c1_axis(ceil((c12-c1)/step):end), c12 - c1_axis(ceil((c12-c1)/step):end), 'LineWidth',2)
axis([ 0 1.1*max([c2 c1]) 0 1.1*max([c2 c1]) ])
xlabel('R_2 (bits/ch. use)')
ylabel('R_1 (bits/ch. use)')
legend('Optimization Result', 'MAC Capacity Region')
grid on
hold off